use std::sync::Arc;

use axum::{
    extract::{Path, State},
    response::Redirect,
    routing::get,
    Router,
};
use sqlx::{query, PgPool};
use tracing::{error, field, instrument};
use ulid::Ulid;
use url::Url;
use uuid::Uuid;

use crate::error::AppError;

#[derive(Clone)]
struct SharedState {
    db: PgPool,
    default_destination: Arc<Url>,
}

pub fn router(db: PgPool, default_destination: Url) -> Router {
    Router::new()
        .route("/:slug", get(redirect))
        .with_state(SharedState {
            db,
            default_destination: Arc::new(default_destination),
        })
}

async fn redirect(
    State(SharedState {
        db,
        default_destination,
    }): State<SharedState>,
    Path(slug): Path<String>,
) -> Result<Redirect, AppError> {
    let result = query!("SELECT id, destination FROM link WHERE slug = $1", slug)
        .fetch_optional(&db)
        .await?
        .map(|r| (Ulid::from(r.id), r.destination));

    Ok(match result {
        Some((id, destination)) => {
            tokio::spawn(increase_visit_count(id, db));
            Redirect::temporary(&destination)
        }
        None => Redirect::temporary(default_destination.as_str()),
    })
}

#[instrument(skip(db))]
async fn increase_visit_count(id: Ulid, db: PgPool) {
    let result = query!(
        "UPDATE link SET visit_count = visit_count + 1 WHERE id = $1",
        Uuid::from(id),
    )
    .execute(&db)
    .await;

    match result {
        Ok(result) if result.rows_affected() != 1 => {
            error!(err = field::display(AppError::ImpossibleAffectedRows(result.rows_affected())));
        }
        Err(err) => error!(err = field::display(err)),
        _ => {}
    }
}
